# -*- coding: utf-8 -*-
"""
"""

import numpy as np
from cslug import ptr

from motmot._slug import slug


class Queue(object):
    """Wrapper class around the C Queue structure.

    For problems that intuitively would be solved by recursive functions but,
    due to the high recursion depths that may involve ``O(len(mesh))`` or
    ``O(len(mesh) ** 2)`` etc., will cause ``StackOverflowErrors`` (the C
    equivalent of ``RecursionError``).

    Instead of invoking itself functions should treat a :meth:`Queue` as a to-do
    list by pop-reading from the front, appending to the end, and stopping when
    the queue is empty.

    This Queue is implemented in C to be used in C. Whilst its methods have been
    ported into Python, they are for debugging only and will likely be slower
    than plain Python ``list``.

    Supported methods in C are:

    .. code-block:: C

        // Add x to the end of the queue if x is not already in it.
        Q_add(queue, x);

        // Append x to the end of the queue - note dangerous. See below.
        Q_append(queue, x);

        // Read and pop the element at the front of the queue.
        int x = Q_consume(queue);

        // Get number of unprocessed elements left in the queue.
        int count = Q_len(queue);

        // Check if queue is empty.
        bool empty = Q_is_empty(queue);

        // Just `append` wrapped in a for loop.
        Q_appends(queue, array, len_array);

    In Python, functions are available as queue methods
    e.g. ``queue.append(x)``.


    This class also makes some tight assumptions and restrictions on usage.

    - Queue elements must be of C ``int`` type and  satisfy
      ``0 <= x < queue.max_size``.

    - It is a strict FIFO (first in first out). Elements can only be added to
      end and removed from the start.

    - Elements in the queue **must be unique**. Use ``add(queue, x)`` instead of
      ``append(queue, x)`` to enforce this.

    - ``queue.max_size`` must be 1 greater than the maximum number of unique
      elements that will go in the queue.

    It is recommended to create this object in Python then pass
    ``queue.c_queue_ptr`` to any C function that needs a queue, rather than
    allocate and (likely forget to) deallocate memory in C.

    """

    def __init__(self, max_size):
        self.queue = np.empty(max_size, dtype=np.intp, order="C")
        self.reverse_queue = np.full_like(self.queue, -1)
        self.max_size = max_size

        self._queue = ptr(self.queue)
        self._reverse_queue = ptr(self.reverse_queue)

        self._raw = slug.dll.Queue(self._queue, self._reverse_queue, 0, 0,
                                   self.max_size)

        self.needs_clearing = None

    @property
    def consume_index(self):
        return self._raw.consume_index

    @property
    def append_index(self):
        return self._raw.append_index

    def append(self, value):
        slug.dll.Q_append(self._raw._ptr, value)

    def appends(self, value):
        value = np.asarray(value, dtype=np.intp, order="C")
        slug.dll.Q_appends(self._raw._ptr, ptr(value), value.size)

    def consume(self):
        return slug.dll.Q_consume(self._raw._ptr)

    def add(self, value):
        slug.dll.Q_add(self._raw._ptr, value)

    def __len__(self):
        return slug.dll.Q_len(self._raw._ptr)

    def __del__(self):
        try:
            slug.dll.Q_destroy_queue(self._raw._ptr)
        except AttributeError:
            pass

    def get_between(self, start, end):
        start %= self.max_size
        end %= self.max_size

        parts = []
        if start > end:
            parts.append(self.queue[start:])
        parts.append(self.queue[start:end])
        if start > end:
            parts.append(self.queue[:end])

        return np.concatenate(parts)

    @property
    def in_waiting(self):
        return self.get_between(self.consume_index, self.append_index)

    def __repr__(self):
        out = [type(self).__name__, "["]
        toggle = self.append_index < self.consume_index
        for (i, x) in enumerate(self.queue):
            if (self.consume_index <= i) == (i < self.append_index) ^ toggle:
                out.append(str(x))
            else:
                out.append("-")
        out += ["]"]
        return " ".join(out)

    def __bool__(self):
        return not slug.dll.Q_is_empty(self._raw._ptr)



# -*- coding: utf-8 -*-
"""
"""

# This is a backport of functools.cached_property().

_UNINITIALISED = object()


class cached_property:  # pragma: no cover

    def __init__(self, func):
        self.func = func
        self.attrname = None
        self.__doc__ = func.__doc__
        from threading import RLock
        self.lock = RLock()

    def __set_name__(self, owner, name):
        if self.attrname is None:
            self.attrname = name
        elif name != self.attrname:
            raise TypeError(
                "Cannot assign the same cached_property to two different names "
                f"({self.attrname!r} and {name!r}).")

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        if self.attrname is None:
            raise TypeError(
                "Cannot use cached_property instance without calling __set_name__ on it."
            )
        cache = instance.__dict__
        val = cache.get(self.attrname, _UNINITIALISED)
        if val is _UNINITIALISED:
            with self.lock:
                # check if another thread filled cache while we awaited lock
                val = cache.get(self.attrname, _UNINITIALISED)
                if val is _UNINITIALISED:
                    val = self.func(instance)
                    cache[self.attrname] = val
        return val



# -*- coding: utf-8 -*-
"""Dumping ground for random bits and bobs."""

import contextlib
from collections import defaultdict as _default_dict
import os
import io
import numpy as np

from motmot._compat import cached_property as _cached_property

idx = type("IDx", (object,), dict(__getitem__=lambda self, x: x))()


class Independency(_default_dict):
    """Manage invalidation of :class:`cached_property`.

    Most cached properties will need refreshing after calling a method which
    modifies the object. There are exceptions. A mesh's area for example doesn't
    change if it is rotated. Provide a means to bulk invalidate cached
    properties and to exclude properties from invalidation.
    """

    def __init__(self):
        super().__init__(set)
        self.cls = None

    def init(self, cls: type):

        self.cls = cls
        self.cached_properties = {
            i.attrname for i in vars(cls).values()
            if isinstance(i, _cached_property) for cls in self.cls.mro()
        }

    def of(self, *operations):

        def wrap(cached: _cached_property):
            for operation in operations:
                self[operation].add(cached.func.__name__)
            return cached

        return wrap

    def reset_on(self, operation):
        return self._resetter(self.cached_properties - self[operation])

    def _resetter(self, to_invalidate):

        def reset(self):
            for i in to_invalidate:
                self.__dict__.pop(i, None)

        return reset

    def reset_all(self, obj):
        for i in self.cached_properties:
            obj.__dict__.pop(i, None)


@contextlib.contextmanager
def _buffer_open(file: io.IOBase, *args, **kwargs):
    """A dummy context manager to be used in place of open() if the file is
    already a either an open file object or a pseudo file."""
    yield file
    return


def _opener(file):
    """Select an appropriate open() like function for a given file type."""
    if isinstance(file, io.IOBase):
        return _buffer_open

    if not isinstance(file, (str, os.PathLike)):
        from builtins import open
        return open

    suffix = os.path.splitext(file)[1]

    if suffix == ".xz":
        from lzma import open
    elif suffix == ".gz":
        from gzip import open
    elif suffix == ".bz2":
        from bz2 import open
    else:
        from builtins import open

    return open


def open_(file, *args, **kwargs):
    """A drop-in replacement to the builtin open() function.

    * If given an open file handle or pseudo file (e.g. io.BytesIO()) the file
      is returned wrapped in a NO-OP context manager.
    * If given a filename ending in a standard compression format suffix (.gz,
      .xz, .bz2) then a appropriate file wrapper is returned which handles the
      compression or decompression (e.g. lzma.open()),

    Otherwise, it defers to the builtin open() behaviour.

    """
    return _opener(file)(file, *args, **kwargs)


def as_nD(x: np.ndarray, n: int, name: str) -> np.ndarray:
    """Prepend new axes until **x** is **n** dimensional. Raise an error if
    **x** has too many dimensions."""
    if x.ndim > n:
        raise ValueError(
            f"'{name}' has too many dimensions. "
            f"A {n}D array was expected but got an array with shape {x.shape}.")
    while x.ndim < n:
        x = x[np.newaxis]
    return x



# -*- coding: utf-8 -*-
"""Some of the graph-like connectivity algorithms which are typically applied to
meshes but may be used elsewhere are exposed here."""

from ._polygon_walk import connected, group_connected



# -*- coding: utf-8 -*-
"""
"""

import numpy as np
from cslug import ptr
from rockhopper import RaggedArray

from motmot._slug import slug


def make_polygon_map(faces: np.ndarray, id_count: int, use_C=True):
    """Create an array mapping each polygon to its neighbours."""
    # This pivots around the fact that if polygon p0 shares an edge v0->v1 with
    # polygon p1, then p1 contains the edge v1->v0 (hard-coding the assumption
    # that a polygon's vertices are consistently listed in anticlockwise order).
    # So to find a polygon's neighbour on a given edge, we need to find that
    # edge reversed in another polygon.

    n, d = faces.shape

    # Normally, we'd tackle this with an {(v0, v1): polygon_id} dictionary but
    # that is far too slow. Instead, group by v0 into ragged arrays.

    _polygons, _sub_ids = np.divmod(np.arange(faces.size, dtype=np.intp),
                                    faces.shape[1])
    _dest_vertices = faces[_polygons, (1 + _sub_ids) % d]

    polygons, dest_vertices = \
        RaggedArray.groups_by(faces.ravel(), _polygons, _dest_vertices,
                              id_max=id_count)

    # Now ``polygons[v0]`` is a list of polygon ids containing v0.
    # And ``dest_vertices[v0]`` is a list of all the v1s that v0 links to.

    # From here it's just a case of searching ``dest_vertices[v1]`` for v0, then
    # checking in ``polygons[v1]`` for which polygon the v0 corresponds to. This
    # step has been translated into C for speed.

    polygon_map = np.empty(faces.shape, dtype=np.intp, order="C")
    if use_C:
        populate_polygon_map_c(polygon_map, faces, n, d, dest_vertices,
                               polygons)
    else:
        populate_polygon_map_py(polygon_map, faces, n, d, dest_vertices,
                                polygons)

    return polygon_map


def populate_polygon_map_c(polygon_map, faces, n, d, dest_vertices: RaggedArray,
                           polygons: RaggedArray):
    assert faces.dtype == np.intp
    assert faces.flags.c_contiguous
    slug.dll.populate_polygon_map(ptr(polygon_map), ptr(faces), n, d,
                                  dest_vertices._c_struct._ptr,
                                  polygons._c_struct._ptr)


def populate_polygon_map_py(polygon_map, faces, n, d, dest_vertices, polygons):
    """A pure-Python proof-of-concept/test for :meth:`populate_polygon_map_c`.
    """
    for i in range(n):
        for j in range(d):
            v0 = faces[i, j]
            v1 = faces[i, (j + 1) % d]
            for k in range(dest_vertices.starts[v1], dest_vertices.ends[v1]):
                if dest_vertices.flat[k] == v0:
                    # Note that ``polygons.starts`` is ``dest_vertices.starts``
                    # and similarly for ``xxx.ends``, which is why it is ok
                    # to use indices ``k`` for ``dest_vertices`` with
                    # ``polygons`` without any kind of conversion or remapping.
                    polygon_map[i, j] = polygons.flat[k]
                    break
            else:
                polygon_map[i, j] = -1


if __name__ == "__main__":
    faces = np.array([[0, 1, 2, 5], [3, 4, 0, 5], [2, 1, 0, 6]], dtype=np.intp)
    id_count = faces.max() + 1
    target = np.array([[2, 2, -1, 1], [-1, -1, 0, -1], [0, 0, -1, -1]])
    assert np.array_equal(make_polygon_map(faces, id_count, use_C=True), target)



# -*- coding: utf-8 -*-
"""
"""
import numpy as np

class UnitVector(np.ndarray):
    def __init__(self, vector): ...
    vector: np.ndarray
    def __call__(self, vector, keepdims=False): ...
    inner_product = __call__
    def matched_sign(self, vector): ...
    def __neg__(self) -> 'UnitVector': ...
    def remove_component(self, vector): ...
    def get_component(self, vector): ...
    def furthest(self, points, n=None, return_projection=False,
                 return_args=False): ...
    def with_(self, point, projection): ...
    def match(self, point, target_point): ...



# -*- coding: utf-8 -*-
"""
"""
from ._version import __version__, __version_info__
from . import geometry
from ._mesh import Mesh
from ._curvatures import Curvature
from . import connectivity


def _PyInstaller_hook_dir():
    import os
    return [os.path.dirname(__file__)]



# -*- coding: utf-8 -*-
"""
Hook for PyInstaller.
"""

from motmot._slug import slug

# Put the cslug binary and its types json in a `motmot` directory.
datas = [(str(slug.path), "motmot"), (str(slug.types_map.json_path), "motmot")]



# -*- coding: utf-8 -*-
"""
"""

import os
from pathlib import Path
import io
import warnings
from typing import Optional, Union
import copy
from numbers import Integral
import operator

import numpy as np
import numpy
from stl.mesh import Mesh as _Mesh
from hirola import HashTable
from rockhopper import RaggedArray

from motmot._compat import cached_property
from motmot._misc import idx, Independency, as_nD, open_
from motmot import geometry


def _subsample(name, indices, doc=""):
    """Create an alias property for slices of mesh data."""

    def get(self):
        return self.vectors[indices]

    def set(self, value):
        self.vectors[indices] = value

    get.__name__ = name
    return property(get, set, None, doc)


independent = Independency()


class Mesh(object):
    __vectors__: np.ndarray
    __faces__: np.ndarray
    __vertices__: np.ndarray
    is_faces_mesh: bool = False
    """If true, this mesh internally uses :attr:`vertices` and :attr:`faces`.
    Otherwise, it uses :attr:`vectors`."""

    def __init__(self, vertices, faces=None, name=""):
        """A :class:`Mesh` can be constructed in three different ways:

        1. From a filename or file-like binary stream::

            mesh = Mesh("my-mesh.stl")

        2. From a 3D :attr:`vectors` array. It should have shape :py:`(n, d, 3)`
           where ``n`` is the number of polygons and ``d`` is the number of
           vertices per polygon.

        3. From :attr:`vertices` and polygon :attr:`faces`.

        Currently only STL files can be read directly. For other 3D files, read
        using meshio_ then convert to :class:`motmot.Mesh`::

            import meshio
            _mesh = meshio.Mesh.read("my-mesh.ply")
            mesh = Mesh(_mesh.points, _mesh.cells[0].data)

        Beware however that this assumes that the mesh uses a fixed number of
        vertices per polygon. :mod:`motmot` doesn't support mixed polygon types.

        .. _meshio: https://github.com/nschloe/meshio

        """
        self.name = ""

        if faces is None:
            if not isinstance(vertices, (list, np.ndarray)):
                with open_(vertices, "rb") as f:
                    # numpy-stl's detection for streams has some holes in it.
                    # For some reason, just passing the open compressed file to
                    # numpy-stl causes it to only read some of it. Create a
                    # redundant intermediate io.BytesIO(). Note that even ASCII
                    # STLs must be read in binary mode.
                    fh = io.BytesIO(f.read())

                mesh = _Mesh.from_file(None, fh=fh, calculate_normals=False)
                self.vectors = mesh.vectors
                if isinstance(vertices, (str, os.PathLike)):
                    self.path = vertices
                self.name = mesh.name
            else:
                self.vectors = vertices
        else:
            self.is_faces_mesh = True
            self.vertices = vertices
            self.faces = faces
        self.name = name or self.name

        self._bounds = np.empty((2, 3), self.dtype)

    _vertex_table: HashTable

    def __repr__(self):
        if not self.is_faces_mesh:
            return f"<Vectors {type(self).__name__} at {hex(id(self))} | " \
                   f"{len(self)} {self.vectors.shape[1]}-sided polygons>"
        return f"<Faces {type(self).__name__} at {hex(id(self))} | " \
               f"{len(self.vertices)} vertices | " \
               f"{len(self)} {self.faces.shape[1]}-sided polygons>"

    @property
    def path(self) -> Optional[Path]:
        """The filename used to open this mesh. This is set to none if this mesh
        was not read directly from a file.

        Returns:
            Either a :class:`pathlib.Path` filename or :py:`None`.

        """
        return getattr(self, "_path", None)

    @path.setter
    def path(self, path):
        self._path = Path(path) if path else None

    @path.deleter
    def path(self):
        self._path = None

    @cached_property
    def vertex_table(self) -> HashTable:
        """The lookup table behind vertex uniquifying and fast vertex lookup.

        This object, a :class:`hirola.HashTable`, is similar to a :class:`dict`
        with this mesh's unique vertices as its keys and an enumeration as its
        values.

        To get a vertex ID (or IDs) for a given point(s) use::

            ids = mesh.vertex_table[points]

        This is the reciprocal of::

            points = mesh.vertices[ids]

        To quickly test if a vertex or vertices is in :attr:`vertices` use
        :meth:`mesh.vertex_table.contains() <hirola.HashTable.contains>`.

        """
        if not self.is_faces_mesh:
            points = self.__vectors__
            length = points.shape[0] * points.shape[1]
        else:
            points = self.__vertices__
            length = len(points)

        # Create a hash table with enough space to fit all points (assuming the
        # worst case where all points are unique) plus 25% padding for speed,
        # and a key dtype of 3 numbers (probably floats).
        table = HashTable(int(1.25 * length), points.dtype * 3)
        # Add all points to the table logging the vertex ID for each point.
        faces = table.add(points)

        if not self.is_faces_mesh:
            # For vectors meshes, set faces attribute.
            self.__faces__ = faces
        elif len(table) < len(self.__vertices__):
            # For faces meshes, `self.faces` should already be set and be
            # identical to the `faces` found above unless there were duplicates
            # in `self.vertices`. These can cause some algorithms to break.
            warnings.warn("Duplicate vertices in mesh.vertices.")

        return table

    @property
    def dtype(self):
        """The :class:`numpy.dtype` of :attr:`vertices` and :attr:`vectors`.
        """
        if self.is_faces_mesh:
            return self.__vertices__.dtype
        return self.__vectors__.dtype

    @property
    def vertices(self) -> np.ndarray:
        """All points in the mesh with duplicity removed.

        If this mesh is not originally an faces mesh, i.e. :attr:`vertices` had
        to be calculated from :attr:`vectors`, then this array is read-only.

        """
        if self.is_faces_mesh:
            return self.__vertices__
        return self.vertex_table.keys

    @vertices.setter
    def vertices(self, x):
        if not self.is_faces_mesh:
            raise ValueError("A vectors mesh's vertices are readonly. "
                             "Write to mesh.vectors instead.")
        # Vertices must be C contiguous for hirola.HashTable.
        x = np.ascontiguousarray(x)

        # If given an empty array:
        if x.size == 0:
            # Make it valid by default.
            x = np.empty((0, 3), x.dtype)

        elif x.shape[-1] != 3:
            raise ValueError("The last axis of vertices must be of length 3. "
                             f"Received an array with shape {x.shape}.")

        # Implicitly promote a single vertex [x, y, z] to [[x, y, z]]
        self.__vertices__ = as_nD(x, 2, "vertices")
        self.reset()

    @property
    def faces(self) -> np.ndarray:
        """Indices of vertices used to construct each polygon.

        Returns:
            Integer array with shape :py:`(len(mesh), mesh.per_polygon)`.

        """
        if self.is_faces_mesh:
            return self.__faces__
        return self._faces_from_vectors

    @faces.setter
    def faces(self, x):
        if not self.is_faces_mesh:
            raise ValueError("A vectors mesh's faces are readonly. "
                             "Write to mesh.vectors instead.")

        # faces must be C contiguous and of a fixed dtype for the various C
        # functions to work.
        x = np.asarray(x, dtype=np.intp, order="C")

        # Set empty input `[]` to something sane. Namely 0 triangles.
        if x.size == 0 and x.ndim == 1:
            x = np.empty((0, 3), np.intp)

        self.__faces__ = as_nD(x, 2, "faces")
        self.reset()

    @independent.of("translate", "rotate")
    @cached_property
    def _faces_from_vectors(self):
        self.vertex_table
        return self.__faces__

    @cached_property
    def _vectors_from_faces(self):
        assert self.is_faces_mesh
        return self.__vertices__[self.__faces__]

    @property
    def vectors(self) -> np.ndarray:
        """The :py:`(x, y, z)` coordinates of each corner of each polygon.

        Returns:
            An :py:`(number of polygons, mesh.per_polygon, 3)` shaped array.

        """
        if self.is_faces_mesh:
            return self._vectors_from_faces
        return self.__vectors__

    @vectors.setter
    def vectors(self, x):
        if self.is_faces_mesh:
            raise ValueError("A faces mesh's vectors are readonly. "
                             "Write to mesh.vertices instead.")
        x = np.ascontiguousarray(x)
        if x.ndim != 3 or x.shape[-1] != 3:
            raise ValueError(
                "Vectors must be a 3D array with last axis of length 3. "
                f"Received an array with shape {x.shape}.")
        self.__vectors__ = x
        self.reset()
        self._bounds = np.empty((2, 3), self.__vectors__.dtype)

    x = _subsample(
        "x", idx[:, :, 0], "The x coordinate of each vertex of each polygon. "
        "Equivalent to :py:`mesh.vectors[:, :, 0]`.")
    y = _subsample(
        "y", idx[:, :, 1], "The y coordinate of each vertex of each polygon. "
        "Equivalent to :py:`mesh.vectors[:, :, 1]`.")
    z = _subsample(
        "z", idx[:, :, 2], "The z coordinate of each vertex of each polygon. "
        "Equivalent to :py:`mesh.vectors[:, :, 2]`.")
    v0 = _subsample(
        "v0", idx[:, 0], "The 1\\ :superscript:`st` corner of each polygon. "
        "Equivalent to :py:`mesh.vectors[:, 0]`.")
    v1 = _subsample(
        "v1", idx[:, 1], "The 2\\ :superscript:`nd` corner of each polygon. "
        "Equivalent to :py:`mesh.vectors[:, 1]`.")
    v2 = _subsample(
        "v2", idx[:, 2], "The 3\\ :superscript:`rd` corner of each polygon. "
        "Equivalent to :py:`mesh.vectors[:, 2]`.")

    def __len__(self):
        """Length is defined as the number of polygons."""
        return len(self.__faces__ if self.is_faces_mesh else self.__vectors__)

    @property
    def per_polygon(self) -> int:
        """The number of corners each polygon has."""
        return (self.__faces__
                if self.is_faces_mesh else self.__vectors__).shape[1]

    @independent.of("translate", "rotate")
    @cached_property
    def vertex_counts(self) -> np.ndarray:
        """The number of times each vertex id appears in :attr:`faces`.

        Returns:
            1D integer array with the same length as :attr:`vertices`.

        """
        counts = np.zeros(len(self.vertices), np.intc)
        np.add.at(counts, self.faces, 1)
        return counts

    @cached_property
    def max(self) -> np.ndarray:
        """The maximum ``x``, ``y`` and ``z`` value. Shares memory with
        :attr:`bounds`."""
        self._bounds[1] = [i.max() for i in (self.x, self.y, self.z)]
        return self._bounds[1]

    @cached_property
    def min(self) -> np.ndarray:
        """The minimum ``x``, ``y`` and ``z`` value. Shares memory with
        :attr:`bounds`."""
        self._bounds[0] = [i.min() for i in (self.x, self.y, self.z)]
        return self._bounds[0]

    @independent.of("translate")
    @cached_property
    def dims(self) -> np.ndarray:
        """The overall length, width and height of the mesh. Or the difference
        between :attr:`max` and :attr:`min`."""
        return self.max - self.min

    @property
    def bounds(self) -> np.ndarray:
        """The minimum and maximum ``(x, y, z)`` values. Equivalent to
        :py:`array([mesh.min, mesh.max])`.

        Returns:
            2D array with shape :py:`(2, 3)`.

        """
        self.min
        self.max
        return self._bounds

    @independent.of("translate")
    @cached_property
    def normals(self) -> np.ndarray:
        """Normals to each polygon.

        Returns:
            2D array with shape :py:`(len(mesh), 3)`.

        Normals point outwards provided that the polygons corners are listed in
        counter-clockwise order (which is the usual convention).

        """
        return np.cross(self.v0 - self.v1, self.v1 - self.v2)

    @independent.of("transform")
    @cached_property
    def areas(self) -> np.ndarray:
        """Surface area of each polygon.

        Returns:
            1D array with length :py:`len(mesh)`.

        Note that for non co-planer (not flat) polygons, this definition becomes
        progressively more arbitrary as the polygons become more complicated.
        Different areas can be obtained for identical polygons simply by
        *rolling* a polygon's corners so that a different corner is listed
        first.

        """
        if self.per_polygon == 3:
            # Triangles are much simpler than other shapes and can therefore
            # take the simpler/faster approach.
            return geometry.magnitude(self.normals) / 2
        return geometry.area(self.vectors)

    @independent.of("translate", "rotate")
    @cached_property
    def area(self) -> float:
        """The total surface area. This is simply the sum of :attr:`areas`."""
        return np.sum(self.areas)

    @independent.of("translate")
    @cached_property
    def units(self):
        """Normalised outward :attr:`normals` for each polygon.

        Returns:
            2D array with shape :py:`(len(mesh), 3)`.

        """
        old = np.seterr(invalid="ignore", divide="ignore")
        if self.per_polygon == 3:
            # Again, this is marginally faster but only works for triangular
            # meshes.
            out = self.normals / (self.areas[:, np.newaxis] * 2)
        else:
            out = geometry.normalised(self.normals)
        np.seterr(**old)
        return out

    @independent.of("translate")
    @cached_property
    def vertex_normals(self) -> np.ndarray:
        """Weighted outward normal for each vertex in :attr:`vertices`.

        Returns:
            2D array with shape :py:`(len(mesh.vertices), 3)`.

        Computed as the normalised average of the surface :attr:`normals` of the
        faces that contain that vertex. Averages are weighted by :attr:`areas`.

        """
        old = np.seterr(invalid="ignore", divide="ignore")
        normals = geometry.normalised(self._vertex_normals)
        np.seterr(**old)
        return normals

    @property
    def _vertex_normals(self):
        """Raw un-normalised vertex normals."""
        normals = np.zeros_like(self.vertices)
        for faces in self.faces.T:
            np.add.at(normals, faces, self.normals)
        return normals

    def translate(self, translation):
        """Move this mesh without rotating."""
        # Avoid inplace array modification because it either breaks or loses
        # precision if the dtypes don't match.
        if self.is_faces_mesh:
            self.__vertices__ = self.__vertices__ + translation
        else:
            self.__vectors__ = self.__vectors__ + translation

        # noinspection PyUnresolvedReferences
        self._reset_on_translate()

    def rotate_using_matrix(self, rotation_matrix, point=None):
        """Rotate inplace, the mesh using a **rotation_matrix**.

        Internally this is just a matrix multiplication where the mesh's
        vertices are *post-multiplied* by **rotation_matrix**.

        """
        if point is not None:
            point = np.asarray(point)
            self.translate(-point)
            self.rotate_using_matrix(rotation_matrix)
            self.translate(point)
            return

        # Inplace matrix multiplication (i.e. ``@=``) is not allowed.
        if self.is_faces_mesh:
            self.__vertices__ = self.__vertices__ @ rotation_matrix
        else:
            self.__vectors__ = self.__vectors__ @ rotation_matrix

        # noinspection PyUnresolvedReferences
        self._reset_on_rotate()

    # Shamelessly nick this from numpy-stl.
    rotation_matrix = staticmethod(_Mesh.rotation_matrix)
    rotate = _Mesh.rotate

    def reset(self):
        """Invalidate all cached properties.

        Use after directly writing or setting one of this mesh's array
        attributes.
        """
        # noinspection PyUnresolvedReferences
        self._reset_all(self)

    def __getitem__(self, item) -> 'Mesh':
        if self.is_faces_mesh:
            return type(self)(self.vertices, self.faces[item], name=self.name)
        return type(self)(self.vectors[item], name=self.name)

    @staticmethod
    def __array__():
        raise TypeError(
            "Meshes can't be converted directly to arrays. Use one "
            "of `vertices`, `vectors` or `faces` attributes instead.")

    def crop(self, mask: Union[np.ndarray, slice],
             in_place: bool = False) -> 'Mesh':
        """Return a subsample of the original mesh. Inclusion is defined by
        **mask**.

        Args:
            mask:
                Polygons to include.
            in_place:
                Modify this mesh instead of making a modified copy, defaults to
                False.
        Returns:
            This mesh if **in_place** or a new cropped one.

        A minimal usage example:

        .. code-block:: python

            # Get only polygons with non-negative average Z-values.
            cropped = mesh.crop(mesh.centers[:, 2] >= 0)

        For a faces mesh this samples :attr:`faces` and leaves
        :attr:`vertices` untouched without copying and is equivalent to::

            cropped = Mesh(mesh.vertices, mesh.faces[mask], name=mesh.name)

        For a vectors based mesh this function simply samples :attr:`vectors`::

            cropped = Mesh(mesh.vectors[mask], name=mesh.name)

        Please ensure you are aware of `when indexing copies in numpy`_ if you
        intend to modify either the cropped or the original mesh afterwards.

        .. _`when indexing copies in numpy`: https://numpy.org/doc/stable/reference/arrays.indexing.html#advanced-indexing

        When **inplace** is false (default), cropping can also be achieved by
        indexing the mesh directly::

            cropped = mesh[mask]

        """
        if not in_place:
            return self[mask]
        if self.is_faces_mesh:
            self.__faces__ = self.__faces__[mask]
        else:
            self.__vectors__ = self.__vectors__[mask]
        self.reset()
        return self

    def __copy__(self):
        cls = type(self)
        out = cls.__new__(cls)
        out.__setstate__(self.__getstate__())
        return out

    def __deepcopy__(self, memo):
        cls = type(self)
        out = cls.__new__(cls)
        state = self.__getstate__()
        state = {key: copy.deepcopy(val, memo) for (key, val) in state.items()}
        out.__setstate__(state)
        return out

    def copy(self, deep=True) -> 'Mesh':
        """Make a shallow or deep copy of the mesh.

        Args:
            deep:
                If true, copy the underlying :attr:`vectors` or :attr:`vertices`
                and :attr:`faces` arrays. Otherwise output will share these
                arrays with this mesh.
        Returns:
            Another mesh.

        Caches of cached properties are never copied.

        """
        return self.__deepcopy__({}) if deep else self.__copy__()

    def __getstate__(self):
        if self.is_faces_mesh:
            return {
                "vertices": self.__vertices__,
                "faces": self.__faces__,
                "name": self.name,
                "path": self.path,
            }
        return {"vectors": self.vectors, "name": self.name, "path": self.path}

    def __setstate__(self, dic):
        self.__init__(dic.get("vectors", dic.get("vertices")), dic.get("faces"),
                      dic["name"])
        self.path = dic["path"]

    @cached_property
    def centers(self) -> np.ndarray:
        """The center of each polygon, defined as the mean of each polygon's
        corners.

        Returns:
            2D array with shape :py:`(len(mesh), 3)`.

        """
        # This could just be `self.vectors.mean(axis=1)` but
        # numpy.ufunc.reduce() with axis != None is much slower than pure
        # Python.
        out = self.vectors[:, 0]
        for i in range(1, self.vectors.shape[1]):
            out = out + self.vectors[:, i]
        return out / self.vectors.shape[1]

    @independent.of("translate", "rotate")
    @cached_property
    def polygon_map(self) -> np.ndarray:
        """Maps each polygon to its adjacent (shares a common edge) polygons.

        The format is an numpy int array with the same shape as
        :attr:`Mesh.faces`. A polygon is referenced by its position in
        :attr:`Mesh.faces` (or :attr:`Mesh.vectors`).

        For example, assume a triangular mesh is called  ``mesh``. And suppose
        :py:`mesh.polygon_map[n]` is :py:`[i, j, k]`, then:

        - The edge going from vertex *0* to vertex *1* of the triangle
          :py:`mesh.vectors[n]` would be shared with the triangle
          :py:`mesh.vectors[i]`,
        - The edge going from vertex *1* to vertex *2* of the triangle
          :py:`mesh.vectors[n]` would be shared with the triangle
          :py:`mesh.vectors[j]`,
        - And the edge going from vertex *2* to vertex **0** of the triangle
          :py:`mesh.vectors[n]` would be shared with the triangle
          :py:`mesh.vectors[k]`,

        Any polygons which are missing a neighbour on a particular edge (i.e.
        on the boundary of a non-closed mesh) use :py:`-1` as a placeholder.

        """
        from motmot._polygon_map import make_polygon_map
        return make_polygon_map(self.faces, len(self.vertices))

    def connected_polygons(self, initial, mask=None, polygon_map=None):
        """Recursively walk connected polygons.

        Finds the whole connected region containing the triangle `arg` where
        `whole connected region` here means that all triangles within that region
        are joined indirectly by at least one triangle.

        Args:
            initial (int or numpy.ndarray):
                The polygon id(s) to start at.
            mask (numpy.ndarray):
                Only include regions covered by `mask` if specified.
            polygon_map (numpy.ndarray):
                An alternative polygon map to use. Defaults to
                :attr:`polygon_map`.
        Returns:
            numpy.ndarray:
                A connected region as a 1D bool array.

        This is implemented by navigating the :attr:`polygon_map`.
        Restrictions on connectivity can be given either with the **mask**
        parameter, which mimics removing polygons, or by overriding
        **polygon_map** and replacing values with ``-1`` to block edges between
        polygons. Note that the **mask** only prevents the algorithm from
        traversing **onto** unmasked polygons. If a polygon is in **initial**
        then it will still be included in the output.

        """
        from motmot._polygon_walk import connected
        polygon_map = self.polygon_map if polygon_map is None else polygon_map
        return connected(polygon_map, initial, mask)

    def group_connected_polygons(self, mask=None, polygon_map=None):
        """Group and enumerate all polygons which are indirectly connected.

        Returns:
            (numpy.ndarray, int):
                A :py:`(group_ids, group_count)` pair.

        Functionally, this is equivalent to calling :meth:`connected_polygons`
        repeatedly until every polygon is assigned a group.

        To convert the output to a list of meshes use::

            from rockhopper import RaggedArray
            ragged = RaggedArray.group_by(mesh.vectors, *mesh.group_connected_polygons())
            sub_meshes = [Mesh(i) for i in ragged]

        Or if your using vertices/faces meshes::

            from rockhopper import RaggedArray
            ragged = RaggedArray.group_by(mesh.faces, *mesh.group_connected_polygons())
            sub_meshes = [Mesh(mesh.vertices, faces) for faces in ragged]

        Note that both will work irregardless of :attr:`is_faces_mesh`, however
        the mismatched implementation will be slower.

        """
        from motmot._polygon_walk import group_connected
        polygon_map = self.polygon_map if polygon_map is None else polygon_map
        return group_connected(polygon_map, mask)

    @cached_property
    def displacements(self) -> np.ndarray:
        """The displacement from each polygon's center to each of its
        neighbours' centers.

        Returns:
            A :py:`(len(mesh), mesh.per_polygon, 3)` numpy array.

        Defaults to ``nan`` when a neighbour is missing.

        """
        return np.where(
            (self.polygon_map != -1)[:, :, np.newaxis],
            self.centers[self.polygon_map] - self.centers[:, np.newaxis],
            np.array([[[np.nan, np.nan, np.nan]]], self.centers.dtype),
        )

    @cached_property
    def curvature(self):
        """Everything curvature related. See :class:`motmot.Curvature`.
        Different *flavours* of curvature are accessible via different
        sub-attributes of this property such as
        :attr:`mesh.curvature.scaleless <Curvature.scaleless>`.
        """
        from motmot._curvatures import Curvature
        return Curvature(self)

    @independent.of("rotate", "translate")
    @cached_property
    def vertex_map(self) -> RaggedArray:
        """A mapping from each vertex id to every other vertex that it is
        directly connected to.

        Each row in this :class:`~rockhopper.RaggedArray` lists all the
        neighbours of one vertex from :attr:`vertices`.
        e.g. If :py:`mesh.vertex_map[10]` is :py:`[13, 17, 19, 22]` then that
        would imply that :py:`mesh.vertices[10]` is connected to each of
        :py:`mesh.vertices[[13, 17, 19, 22]]` by a single polygon edge.

        This mapping is guaranteed to:

        * Be symmetric. If vertex **A** is connected to vertex **B** then **B**
          is connected to **A**.
        * Contain no self references. **A** will never be listed as connected to
          **A**.
        * Contains no duplicates. **A** will never be listed as connected to
          **B** twice.

        The order in which neighbours appear is arbitrary and not guaranteed to
        be consistent across :mod:`motmot` versions.

        .. seealso::

            :meth:`connected_vertices` if you prefer to work directly with
            vertices rather than vertex faces.

        """
        # In a closed mesh each edge is written once going from A to B and once
        # going from B to A.
        # In a non closed mesh, edges on boundaries will only be listed one way.
        # This function must make sure to mirror the *singular* edges on
        # boundaries (obviously without accidentally duplicating any others).

        # Work out which (if any) edges will need to be copied backwards.
        singular = np.argwhere(self.polygon_map == -1)

        # To make the map requires an array listing the start of each edge
        # and an array listing the end of each edge.
        starts = np.empty(self.faces.size + len(singular), np.intc)
        ends = np.empty(self.faces.size + len(singular), np.intc)

        for i in range(self.per_polygon):
            # Add only the counter-clockwise edges.
            # There is usually no need to add clockwise edges because the
            # polygon on the opposite side of the edge will add it.
            s = slice(i * len(self), (1 + i) * len(self))
            starts[s] = self.faces[:, i]
            ends[s] = self.faces[:, (i + 1) % self.per_polygon]

        if len(singular):
            # Add the edges going clockwise for the *singular* edges. i.e
            # Those which have no polygon on the other side of the edge.
            s = slice(-len(singular), None)
            starts[s] = self.faces[singular[:, 0],
                                   (singular[:, 1] + 1) % self.per_polygon]
            ends[s] = self.faces[singular[:, 0], singular[:, 1]]

        return RaggedArray.group_by(ends, starts, len(self.vertices))

    def connected_vertices(self, vertex: np.ndarray) -> np.ndarray:
        """List vertices which are directly connected to **vertex** by one
        polygon edge.

        Args:
            vertex:
                A single point from :attr:`vertices`.
                A NumPy array with shape ``(3,)``.
        Returns:
            An :py:`(n, 3)` array where ``n`` is the number of connected
            vertices.
        Raises:
            KeyError:
                If **vertex** is not in :attr:`vertices`.

        .. seealso::

            This method uses :attr:`vertex_map` under the hood.
            Use :attr:`vertex_map` if you're using vertex ids instead of raw
            vertices.

        """
        if vertex.shape != (3,):
            raise NotImplementedError("Must take a single vertex.")
        return self.vertices[self.vertex_map[self.vertex_table[vertex]]]

    @cached_property
    def reverse_faces(self):
        """A mapping of which polygons each vertex is in.

        This mapping uses flat indices. i.e. To find all instances of vertex
        123 use::

            polygon_ids, corners = \\
                np.divmod(self.reverse_ids[123], self.per_polygon)

        Then :py:`self.faces[polygon_ids, corners]` will all equal 123.

        """
        return RaggedArray.group_by(np.arange(self.faces.size),
                                    self.faces.ravel())

    def on_boundary(self, vertex: Union[np.ndarray, Integral]) -> bool:
        """Test if a **vertex** touches the edge of this mesh.

        Args:
            vertex:
                A 3D point in :attr:`vertices`. Or a single vertex ID.
        Returns:
            True if it touches, false otherwise.

        .. note::

            To check if a polygon touches a mesh edge, simply use::

                any(mesh.polygon_map[polygon_id] == -1)

        """
        if isinstance(vertex, Integral):
            id = vertex
        else:
            id = self.vertex_table[vertex]
        if not np.isscalar(id):
            raise ValueError("Only single vertices are supported.")
        polygons, sub_faces = np.divmod(self.reverse_faces[id],
                                        self.per_polygon)
        return np.any(self.polygon_map[polygons, sub_faces] == -1) \
               or np.any(self.polygon_map[polygons, sub_faces - 1] == -1)

    @cached_property
    def kdtree(self):
        """A KDTree_ with :attr:`centers` as its input data.

        This object powers all non-exact point lookup operations such as
        :meth:`closest_point`. Use its ``query_xxx()`` methods for more flexible
        lookup.

        .. _kdtree: https://github.com/storpipfugl/pykdtree

        """
        assert np.isfinite(self.centers).all()
        from pykdtree.kdtree import KDTree
        return KDTree(self.centers)

    def closest_point(self, target,
                      distance_upper_bound: Optional[float] = None,
                      interpolate=True) -> np.ndarray:
        """Find the nearest point on the surface of the mesh to **target**.

        Args:
            target:
                The point(s) to query. A :class:`numpy.ndarray` with
                :py:`shape[-1] == 3`.
            distance_upper_bound:
                A optional maximum allowed distance from **target** before
                giving up. In this case :py:`nan` is returned.
            interpolate:
                If true, the output will be the nearest point on the surface of
                the nearest polygon. Otherwise, it will only be the center of
                the nearest polygon.
        Returns:
            The nearest point(s) on the surface of the mesh. A
            :class:`numpy.ndarray` with the same shape as **target**.

        .. note::

            Under extreme circumstances, namely if **point** lies between two
            very close parallel-ish surfaces with enormous polygons, then the
            output is not guaranteed to be optimal. It is, however, guaranteed
            to be better than querying the nearest vertex.

        """
        target = np.asarray(target, dtype=self.kdtree.data.dtype)

        # PyKDTree only accepts 2D input arrays.
        # For single points (1D array) or arrays of arrays of points (>2D) we
        # need to convert inputs to 2D arrays then convert the outputs back
        # again. Ideally this would be fixed in pykdtree but the project is
        # borderline abandoned.
        shape = target.shape
        target = target.reshape((-1, 3))
        out = np.empty(target.shape, dtype=self.kdtree.data.dtype)

        # Find the nearest polygon.
        if distance_upper_bound is None:
            distances, args = self.kdtree.query(target)
            mask = slice(None)
        else:
            distances, args = self.kdtree.query(
                target, distance_upper_bound=distance_upper_bound)
            mask = np.isfinite(distances)
            out[~mask] = np.nan
            args = args[mask]

        if interpolate:
            out[mask] = geometry.snap_to_plane(target[mask], self.centers[args],
                                               self.units[args])
        else:
            out[mask] = self.centers[args]

        return out.reshape(shape)

    def local_maxima(self, heights: np.ndarray, boundaries: bool = True,
                     strict: bool = True) -> np.ndarray:
        """Find all :attr:`vertices` whose corresponding value in **heights** is
        greater than that of all its :meth:`connected_vertices`.

        Args:
            heights:
                A per-vertex scalar to rank by.
            boundaries:
                If false, ignore any vertices which :meth:`touch the mesh
                boundary <on_boundary>`.
            strict:
                If true, a vertex's value from **heights** must be strictly
                greater than its neighbours. Otherwise, it may be greater or
                equal.
        Returns:
            The vertex ids of the vertices which are local maxima.

        """
        mask = np.ones(len(self.vertices), bool)

        heights = np.asarray(heights)
        if heights.shape != (len(self.vertices),):
            raise ValueError("`heights` must be a 1D array with the same "
                             "length as `vertices`.")

        less_than = operator.lt if strict else operator.le

        for i in range(self.per_polygon):
            this = self.faces[:, i]
            next = self.faces[:, i - 1]
            # If the next vertex round is higher, then this is not a local
            # maxima. Read as:
            #   mask[this] &= less_than(heights[next], heights[this])
            np.logical_and.at(mask, this, less_than(heights[next],
                                                    heights[this]))
            # Strictly speaking, we could skip this line for closed meshes.
            # If this vertex is higher than the next vertex round, then the next
            # vertex is not a local maxima. Read as:
            #   mask[next] &= less_than(heights[this], heights[next])
            np.logical_and.at(mask, next, less_than(heights[this],
                                                    heights[next]))

        args = np.nonzero(mask)[0]
        if not boundaries:
            args = np.array([i for i in args if not self.on_boundary(i)])

        return args

    def save(self, file):
        """Write the mesh to a file or pseudo file. Currently only STL format
        and compressed variants of STL (``.stl.xz``) are supported."""
        data = np.empty(len(self), _Mesh.dtype)
        mesh = _Mesh(data, name=self.name, calculate_normals=False)
        mesh.vectors = self.vectors
        mesh.normals = self.units
        mesh.name = self.name
        # Monkeypatch numpy-stl's header making method to use just the name we
        # gave it.
        mesh.get_header = lambda *args: (mesh.name or "")[:80].ljust(80, " ")
        with open_(file, "wb") as f:
            mesh.save("unused", fh=f, update_normals=False)


independent.init(Mesh)
Mesh._reset_on_rotate = independent.reset_on("rotate")
Mesh._reset_on_translate = independent.reset_on("translate")
Mesh._reset_all = independent.reset_all



# -*- coding: utf-8 -*-
"""
"""

import numpy as np
from cslug import ptr

from motmot._queue import Queue, slug


def connected(polygon_map, initial, mask=None):
    """Find all nodes connected to the **initial** node(s)."""
    # Just a wrapper for :c:`connected()` from ``polygon_map.c``.

    initial = np.asarray(initial, dtype=np.intp, order="C").ravel()
    assert (0 <= initial).all() and (initial < len(polygon_map)).all()

    polygon_map = mask_polygon_map(polygon_map, mask)

    queue = Queue(len(polygon_map) + 1)

    assert polygon_map.dtype == np.intp
    slug.dll.connected(ptr(initial), len(initial), polygon_map.shape[1],
                       ptr(polygon_map), queue._raw._ptr)

    return queue.get_between(0, queue.append_index)


def mask_polygon_map(polygon_map, mask):
    if mask is not None:
        if mask.ndim == 1:
            mask = mask[polygon_map]
        polygon_map = np.where(mask, polygon_map, -1)
    return polygon_map


def group_connected(polygon_map, mask=None):
    """Group all connected nodes."""
    # Wrap :c:`group_connected()` from ``polygon_map.c``.

    polygon_map = mask_polygon_map(polygon_map, mask)
    queue = Queue(len(polygon_map) + 1)

    group_ids = np.full(len(polygon_map), -1, np.intp, order="C")

    groups_count: int
    groups_count = slug.dll.group_connected(ptr(polygon_map),
                                            ptr(polygon_map.ctypes.shape),
                                            ptr(group_ids), queue._raw._ptr)

    return group_ids, groups_count



# -*- coding: utf-8 -*-
"""
"""

from ._geometry import (
    area,
    center_of_mass,
    closest,
    furthest,
    get_components,
    get_components_zipped,
    inner_product,
    magnitude,
    magnitude_sqr,
    normalise,
    normalised,
    orthogonal_bases,
    snap_to_plane,
    unzip,
    zip,
)
from ._geometry import _reduce_last_axis
from ._unit_vector import UnitVector



# -*- coding: utf-8 -*-
"""
"""

import numpy as np

from motmot._compat import cached_property
from motmot.geometry import inner_product, magnitude
from motmot._mesh import Mesh


class Curvature(object):
    """A bucket for everything curvature related.

    Unlike the more traditional definition of curvature, this is defined per
    polygon edge rather than per vertex. Edges are defined as the shared two
    vertices between a polygon and a neighbour from :attr:`Mesh.polygon_map`.

    Curvature can have several similar forms. In most cases you will want
    :attr:`scaleless` or :attr:`signed` for an orientation and resolution
    independent measurement. :attr:`directional` combined with
    :func:`geometry.inner_product` can be useful for searching for very specific
    shapes with a predefined orientation.

    """
    _mesh: Mesh

    def __init__(self, mesh):
        self._mesh = mesh

    @cached_property
    def directional(self) -> np.ndarray:
        """Curvature's rawest form - the cross product of each triangle's unit
        normal with each of its adjacent triangles' unit normals.

        Returns:
            A :py:`(len(mesh), mesh.per_polygon, 3)` numpy array.

        This form gives a vector rather than a scalar value for each edge.
        The magnitude of this vector is the ``sin()`` of the angle between the
        two polygons and the direction is tangential to the edge between them.

        """
        out = np.cross(self._mesh.units[self._mesh.polygon_map],
                       self._mesh.units[:, np.newaxis])

        return np.where(self._mesh.polygon_map[:, :, np.newaxis] != -1, out,
                        np.array([[[np.nan, np.nan, np.nan]]], dtype=out.dtype))

    @cached_property
    def magnitude(self) -> np.ndarray:
        """Curvature magnitudes about polygons' edges.

        Returns:
            A :py:`(len(mesh), mesh.per_polygon)` numpy array.

        Given by taking magnitudes of :attr:`directional`.

        """
        return magnitude(self.directional)

    @cached_property
    def scaleless(self) -> np.ndarray:
        r"""The reciprocal of the radius of curvature. Or in pig's English, a
        sphere with radius ``10`` will have scaleless curvature ``1 / 10 = 0.1``
        throughout.

        Returns:
            A :py:`(len(mesh), mesh.per_polygon)` numpy array.

        This property is independent of mesh resolution. Given by
        :attr:`magnitude` / :attr:`magnitude(displacements)<Mesh.displacements>`.

        """
        old = np.seterr(divide="ignore", invalid="ignore")
        out = self.magnitude / magnitude(self._mesh.displacements)
        np.seterr(**old)
        return out

    @cached_property
    def signed(self) -> np.ndarray:
        """Signed :attr:`scaleless` curvature magnitudes.

        Like :attr:`scaleless` but signed so that bumps or bulges have positive
        values and slots or grooves or creases have negative values.

        """
        signs = np.sign(
            inner_product(
                self._mesh.units[:, np.newaxis],
                self._mesh.displacements,
            )) * -1
        return self.scaleless * signs



# -*- coding: utf-8 -*-
"""
"""

from typing import Tuple

import numpy as np
import numpy  # Needed for PyCharm's type checking.


def inner_product(a, b, keepdims=False):
    """Calculates the scalar/inner/interior/dot/"whatever you want to call it"
    product of vectors **a** and **b**, returning a scalar.

    Arguments **a** and **b** must be numpy-broadcastable.

    .. seealso::

        The :class:`UnitVector` class for a more convenient way to perform
        multiple :meth:`inner_product` calls.
    """
    return _sum_last_axis(np.asarray(a) * b, keepdims=keepdims)


def magnitude(vector, keepdims=False):
    """Calculate the hypotenuse/magnitude/length  of a **vector**."""
    return np.sqrt(magnitude_sqr(vector, keepdims))


def magnitude_sqr(vector, keepdims=False):
    """Calculate the square of the hypotenuse of a **vector**.

    This is faster than :func:`magnitude` because it skips taking the square
    root and can be used to compare or sort distances.

    """
    vector = np.asarray(vector)
    return _sum_last_axis(vector * vector, keepdims)


def _sum_last_axis(x, keepdims=False):
    """Faster ``np.sum(x, axis=-1, keepdims=keepdims)``."""
    return _reduce_last_axis(x, np.add, keepdims)


def _reduce_last_axis(x, operator, keepdims=False):
    """Faster :meth:`numpy.ufunc.reduce()`."""
    if keepdims:
        return _reduce_last_axis(x, operator)[..., np.newaxis]

    x = np.asarray(x)

    if x.shape[-1] == 0:
        return np.full(x.shape[:-1], operator.identity, x.dtype)

    itr = iter(x.T)
    out = next(itr)
    for i in itr:
        out = operator(out, i)
    out = out.T

    return out


def furthest(points, direction, n=None, return_projection=False,
             return_args=False):
    """Select the point the furthest in **direction**.

    Args:
        points (numpy.ndarray):
            Some points.
        direction (numpy.ndarray or UnitVector):
            A direction.
        n (int or None):
             Specify the furthest **n** points instead of just one.
        return_projection (bool):
             If true, also return the projection of the furthest point(s).
        return_args (bool):
            If true, also return the index(es) similar to :func:`numpy.argmax`.

    """
    points = np.asarray(points)

    if points.size == 0:
        raise ValueError(
            "Can't select furthest from an empty array of `points`.")

    heights = inner_product(points, direction)

    if n is None:
        args = np.argmax(heights, axis=None)
    else:
        if not n:
            args = np.empty(0, int)
        else:
            args = np.argpartition(heights, -n, axis=None)[-n:]
            args = args[np.argsort(heights[np.unravel_index(
                args, heights.shape)])][::-1]

    args = np.unravel_index(args, heights.shape)

    out = points[args]
    if not (return_args or return_projection):
        return out
    out = out,
    if return_projection:
        out += (heights[args],)
    if return_args:
        out += (args,)
    return out


def normalise(vector):
    """Modify in-place **vector** so that it has magnitude ``1.0``.

    Args:
        vector (numpy.ndarray):
            Vector(s) to normalise.

    Returns:
        numpy.ndarray:
            The original magnitudes of **vector**.

    .. seealso:: :func:`normalised` which leaves the original as-is.

    """
    magnitudes = magnitude(vector, keepdims=True)
    vector /= magnitudes
    return magnitudes


def normalised(vector):
    """Return a normalised copy of **vector**.

    .. seealso:: :func:`normalise` to modify in-place.

    """
    return vector / magnitude(vector, keepdims=True)


def get_components(points, *unit_vectors):
    """Get the inner product for each unit vector as separate arrays.

    Args:
        points (numpy.ndarray):
        *unit_vectors (numpy.ndarray or UnitVector):

    Returns:
        tuple[numpy.ndarray]:
            Projections for each unit vector.

    """
    return tuple(inner_product(points, uv) for uv in unit_vectors)


def get_components_zipped(points, *unit_vectors):
    """Get the inner product for each unit vector.

    Args:
        points:
        *unit_vectors (numpy.ndarray or UnitVector):

    Returns:
        numpy.ndarray:
            Projections as one array.

    The unit-vector iteration corresponds to the last axis of the output. i.e ::

        out[..., i] == inner_product(unit_vectors[i], points)

    The more linear-algebra savvy developer will know that this is just a matrix
    multiplication. This function is purely to reduce confusion for those (like
    me) who can never remember if you pre or post multiply and when to
    transpose.

    """
    return points @ np.array(unit_vectors).T


def center_of_mass(points, weights=None):
    """The (weighted) mean of **points**."""
    if weights is None:
        return np.array([i.mean() for i in points.T])
    else:
        weights = weights[(...,) + (np.newaxis,) * (points.ndim - weights.ndim)]
        return np.array([i.sum() for i in (points * weights).T]) / weights.sum()


def orthogonal_bases(v0):
    """Create a set of perpendicular axes.

    Args:
        v0 (numpy.ndarray):
            The first axis.

    Returns:
        tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]:
            Three perpendicular unit vectors.

    The first returned axis is just the input :func:`normalised`. Although there
    are many possible valid outputs for any given input, this function is fully
    deterministic. Reproducibility is guaranteed.

    """
    old = np.seterr(divide="ignore", invalid="ignore")

    v0 = np.asarray(v0)
    shape = v0.shape
    v0 = v0.reshape((-1, v0.shape[-1]))
    v0 = normalised(v0)
    if not np.isfinite(v0).all():
        raise ValueError("v0 must have non zero magnitude")

    to_do = None
    v1 = np.empty_like(v0)
    v2 = np.empty_like(v0)

    for seed in np.eye(3):  # pragma: no branch
        if to_do is not None and len(to_do) == 0:
            break
        v0[to_do], v1[to_do], v2[to_do], _to_do = \
            _orthogonal_bases(v0[to_do], seed)
        if to_do is None:
            to_do = _to_do
        else:
            to_do = to_do[_to_do]

    np.seterr(**old)
    return v0.reshape(shape), v1.reshape(shape), v2.reshape(shape)


def _orthogonal_bases(v0, seed):
    v1 = np.cross(v0, seed)
    v1_mag = normalise(v1)
    v2 = np.cross(v0, v1)
    return v0, v1, v2, (v1_mag.ravel() == 0).nonzero()[0]


def area(polygon) -> np.ndarray:
    """Calculate the area of arbitrary polygons."""
    polygon: np.ndarray = np.asarray(polygon)
    assert polygon.ndim >= 2
    assert polygon.shape[-1] == 3

    if polygon.shape[-2] < 3:
        # Area of a line or a dot is 0.
        return np.zeros(polygon.shape[:-2], polygon.dtype)

    # Triangulate using the first vertex as a focus. i.e. A hexagon:
    #   (p0, p1, p2, p3, p4, p5)
    # becomes 4 triangles which all start at p0:
    #   (p0, p1, p2)
    #   (p0, p2, p3)
    #   (p0, p3, p4)
    #   (p0, p4, p5)

    from_v0 = polygon[..., 1:, :] - polygon[..., 0, np.newaxis, :]
    crosses = [(np.cross(from_v0[..., i, :], -from_v0[..., i + 1, :]))
               for i in range(polygon.shape[-2] - 2)]

    # Area calculation can be simple but isn't if either all vertices are not
    # co-planer (in which case `area` is technically arbitrary as it differs
    # depending on which vertex is used as a triangulation focus) or if the
    # polygon contains inside corners > 180° (think of a pie with a wedge
    # removed).

    # In the simple case, just magnitude() each item in `crosses`, add them up,
    # then divide by 2.
    # Non-co-planer polygons are doomed almost by definition - this complication
    # is more or less ignored.
    # To handle the 2nd problem, some sign magic must be done to treat the
    # missing slice of the pie as a negative area. This is done by looking at
    # normals to each triangle: The normal to a missing wedge will point in
    # the opposite direction to all the other triangles' normals.

    # Generate a consensus normal to this polygon which can be used to
    # sign-match individual triangle normals and thereby detect missing wedges.
    normal = sum(crosses[1:], crosses[0])

    # Calculate a signed area for each triangle. Note that these areas are
    # doubled what they should be.
    areas_x2 = (
        np.copysign(magnitude(i), inner_product(normal, i)) for i in crosses)

    # Add all double-areas together and un-double them.
    return sum(areas_x2, next(areas_x2)) / 2


def zip(*axes) -> np.ndarray:
    """Combine separate x, y, z arrays into a single *points* array.

    Args:
        axes:
            Each separate axis to combine.
    Returns:
        A single array with :py:`shape[-1] == len(axes)`.

    All `axes` must have the matching or broadcast-able shapes. The number of
    axes doesn't have to be 3.

    .. code-block:: python

        >>> zip(np.arange(5), np.arange(-2, 3))
        array([[ 0, -2],
               [ 1, -1],
               [ 2,  0],
               [ 3,  1],
               [ 4,  2]])

        >>> zip(np.arange(10), 4, np.arange(-5, 5))
        array([[ 0,  4, -5],
               [ 1,  4, -4],
               [ 2,  4, -3],
               [ 3,  4, -2],
               [ 4,  4, -1],
               [ 5,  4,  0],
               [ 6,  4,  1],
               [ 7,  4,  2],
               [ 8,  4,  3],
               [ 9,  4,  4]])

    .. seealso:: :func:`unzip` for the reverse.

    This function is similar to :data:`numpy.c_` except that it is not limited
    to 2D arrays.

    """

    return np.concatenate(
        [i[..., np.newaxis] for i in np.broadcast_arrays(*axes)], axis=-1)


def unzip(points) -> Tuple[np.ndarray, ...]:
    """Separate each component from an array of points.

    Args:
        points: Some points.
    Returns:
        Each axis separately as a tuple.

    This is the inverse of :func:`zip`.

    """
    points = np.asarray(points)
    return tuple(points[..., i] for i in range(points.shape[-1]))


def closest(points, target):
    """Select from **points** the point which is closest to **to**.

    Args:
        points:
            An array of vertices to choose from.
        target:
            A single target vertex to be closest to.
    Returns:
        One vertex from **points**.

    """
    points = np.asarray(points)
    index = magnitude_sqr(points - target).argmin()
    return points[np.unravel_index(index, points.shape[:-1])]


def snap_to_plane(point, origin, normal) -> np.ndarray:
    """Map **point** to its nearest point on a plane.

    Args:
        point:
            A vertex or vertices to move.
        origin:
            The plane's origin. Or any point already on the plane(s).
        normal:
            A unit normal to the plane.
    Returns:
        The translated points.

    Alternatively, you may think of this as move **point** along
    **normal** until the resulting output point satisfies::

        inner_product(normal, output) == inner_product(normal, point)

    """
    point = np.asarray(point)
    normal = np.asarray(normal)
    height_from_plane = inner_product(origin - point, normal, keepdims=True)
    return point + height_from_plane * normal



# -*- coding: utf-8 -*-
"""
"""

from cslug import CSlug, anchor, Header

queue_h = Header(*anchor("queue.h", "queue.c"), includes="_queue.h")
slug = CSlug(anchor("motmot", "queue.c", "_queue.h", "polygon_map.c"),
             headers=queue_h)



# -*- coding: utf-8 -*-
"""
"""

from setuptools import setup, find_packages
import runpy
from pathlib import Path

from cslug.building import (build_slugs, bdist_wheel, CSLUG_SUFFIX,
                            copy_requirements)

HERE = Path(__file__).resolve().parent

readme = (HERE / 'README.md').read_text("utf-8")

setup(
    author="Brénainn Woodsend",
    author_email='bwoodsend@gmail.com',
    python_requires='>=3.7',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    description="A sophisticated mesh class for analysing 3D surfaces.",
    install_requires=copy_requirements(),
    entry_points={"pyinstaller40": "hook-dirs=motmot:_PyInstaller_hook_dir"},
    extras_require={
        "test": [
            'pytest>=3',
            'pytest-order',
            'coverage',
            'pytest-cov',
            'coverage-conditional-plugin',
        ]
    },
    license="MIT license",
    long_description=readme,
    long_description_content_type='text/markdown',
    package_data={
        "motmot": ["*" + CSLUG_SUFFIX, "*.json", "geometry/_unit_vector.pyi"]
    },
    keywords='motmot',
    name='motmot',
    packages=find_packages(include=['motmot', 'motmot.*']),
    url='https://github.com/bwoodsend/motmot',
    version=runpy.run_path(HERE / "motmot/_version.py")["__version__"],
    zip_safe=False,
    cmdclass={
        "build": build_slugs("motmot._slug:slug"),
        "bdist_wheel": bdist_wheel,
    },
)



# -*- coding: utf-8 -*-
"""
"""

import inspect
import re

import numpy as np

from motmot.geometry import inner_product, furthest, normalised


class UnitVector(object):
    """Unit vectors symbolise directions.

    A :class:`UnitVector` wraps around an numpy array, stored in the
    :attr:`vector` attribute, which contains the actual data. It behaves
    exactly like an array but with extra methods.

    It is also callable, which applies the inner-product.

    **Usage Example**:

    Suppose we have a point cloud of ``100`` points called ``points``.

    .. code-block:: python

        import numpy as np
        from motmot import geometry

        points = np.random.uniform(-30, 30, (100, 3))

    In a wonky coordinate system that has been so that `up` is actually the
    diagonal unit-vector below.

    .. math::

        \\begin{bmatrix}
        \\frac{3}{5}\\quad
        \\frac{4}{5}\\quad
        0
        \\end{bmatrix}

    Which is passed to Python using the following. Note that the vector is
    normalised automatically.

    .. code-block:: python

        >>> up = geometry.UnitVector([3, 4, 0])
        >>> up
        UnitVector([0.6 0.8 0. ])

    This can do everything a numpy array can do. Any outputs produced are of
    type :class:`numpy.ndarray`.

    .. code-block:: python

        >>> up * 2
        array([1.2, 1.6, 0. ])
        >>> up + 1
        array([1.6, 1.8, 1. ])
        >>> up + up
        array([1.2, 1.6, 0. ])

    There is one exception to the above: Negating it
    returns another :class:`UnitVector`. ::

        >>> -up
        UnitVector([-0.6 -0.8 -0. ])

    To get the heights of **points** use the inner product. ::

        heights = up.inner_product(points)

    A typical 3D analysis will involve lots of inner-product calls so the above
    gets combersome pretty quickly. For convenience, you may instead call the
    vector directly. ::

        heights = up(points)

    Once you have your heights, the following numpy functions are your friends::

        heights.min()   # minimum height
        heights.max()   # maximum height
        heights.mean()  # average height
        heights.ptp()   # `peak to peak` equivalent to max - min

    Get the highest point using::

        highest = up.furthest(points)

    Or::

        highest, max_height = up.furthest(points, return_projection=True)

    If you actually want the lowest then invert ``up`` to get a ``down`` vector.
    ::

        lowest = (-up).furthest(points)

    Methods :meth:`get_component` and :meth:`remove_component` split a point
    into parts parallel and perpendicular to ``up``.

    Methods :meth:`with_` and :meth:`match` return points
    modified to have specified heights. They both have the effect of mapping
    the input onto a plane with ``up`` as the plane normal.

    .. code-block:: python

        # Get `points` but with a height of 10.
        up.with_projection(points, 10)
        # Get `points` but with the same height as ``points[0]``.
        up.match_projection(points, points[0])

    """

    def __init__(self, vector):
        """
        Args:
            vector(numpy.ndarray or UnitVector or str or list or tuple):

                The direction as a vector, or an axis name as a string.

        **vector** is normalised automatically.

        The str axis name format is an optional sign (+-) followed by any of
        ``'ijkxyz'``. Whitespaces and case are ignored. ::

            >>> UnitVector("Z")
            UnitVector([0. 0. 1.])
            >>> UnitVector("+x")
            UnitVector([1. 0. 0.])
            >>> UnitVector("-y")
            UnitVector([ 0. -1.  0.])
            >>> UnitVector("-j")
            UnitVector([ 0. -1.  0.])

        """
        if isinstance(vector, str):
            vector = _str_to_vector(vector)
        self.vector = normalised(np.asarray(vector))

    vector: np.ndarray
    """The raw numpy vector."""

    def __repr__(self):
        return self.__class__.__name__ + "({})".format(self.vector.tolist())

    def __call__(self, vector, keepdims=False):
        return inner_product(self.vector, vector, keepdims=keepdims)

    inner_product = __call__

    def matched_sign(self, vector):
        """Return a reversed copy of **vector** if ``self(vector) < 0``. i.e if
        the angle between *this* vector and **vector** is more than 180°.
        """
        return vector * np.where(self(vector) < 0, -1, 1)

    def __neg__(self):
        return type(self)(-self.vector)

    def remove_component(self, vector):
        """Get the components of **vector** perpendicular to *this* direction.
        """
        return vector - self.get_component(vector)

    def get_component(self, vector):
        """Get the component of **vector** parallel to *this* direction."""
        return self(vector, keepdims=True) * self.vector

    def furthest(self, points, n=None, return_projection=False,
                 return_args=False):
        """Select, from **points**, the point with the highest projection in
        *this* direction.
        """
        return furthest(points, self.vector, n, return_projection, return_args)

    def with_(self, point, projection):
        """Translate **point** along *this* direction so that
        ``self(point) == projection``.

        The output is a modified copy.
        """
        return (np.asarray(projection)[..., np.newaxis] -
                self(point, keepdims=True)) * self + point

    def match(self, point, target_point):
        """Translate **point** so that ``self(point) == self(target_point)``

        Or the returned point is inline with **target_point**.
        """
        return self.with_(point, self(target_point))


_EXCLUDES = {
    "__array_finalize__",
    "__array_function__",
    "__array_interface__",
    "__array_prepare__",
    "__array_priority__",
    "__array_struct__",
    "__array_ufunc__",
    "__array_wrap__",
    "__neg__",
    "__reduce__",
    "__getattr__",
    "__getattribute__",
    "__class__",
    "__copy__",
    "__deepcopy__",
    "__reduce_ex__",
    "__prepare__",
    "__setstate__",
    "__new__",
    "__setattr__",
    "__setattribute__",
    "__dir__",
}  # | set(dir(UnitVector)) - {"__eq__"}

_METHOD_SOURCE = {}
# _METHOD_SOURCE.update(dict.fromkeys(dir(object), object))
_METHOD_SOURCE.update(
    dict.fromkeys(set(dir(np.ndarray)) - _EXCLUDES, np.ndarray))
_METHOD_SOURCE.update(
    dict.fromkeys(
        set(dir(UnitVector)) -
        set(i for i in dir(UnitVector)
            if getattr(object, i, None) is getattr(UnitVector, i)),
        UnitVector,
    ))


def _unitvector_proxy_wrap(attr):

    def f(self, *args):
        #print(attr)
        return getattr(self.vector, attr)(*args)

    return f


for (_attr, _value) in inspect.getmembers(np.ndarray, callable):
    if _METHOD_SOURCE.get(_attr) is np.ndarray:
        setattr(UnitVector, _attr, _unitvector_proxy_wrap(_attr))

_AXES = dict(zip("ijk", np.eye(3)))
_AXES.update(zip("xyz", np.eye(3)))
_SIGNS = {"+": 1, None: 1, "-": -1}
_str_vector_re = re.compile(r"\s*([+-])?\s*(?:([i-k])|([x-z]))")


def _str_to_vector(x):
    """Convert an axis name, such as 'Z' to a unit vector [0, 0, 1]."""
    match = _str_vector_re.match(x.lower())
    if match is None:
        raise ValueError("{!r} is not a valid unit-vector.".format(x))
    sign, ijk, xyz = match.groups()
    if ijk is None:
        axis = ord(xyz) - ord("x")
    else:
        axis = ord(ijk) - ord("i")
    out = np.zeros(3)
    out[axis] = _SIGNS[sign]
    return out
